# -*- coding: utf-8 -*-

# Copyright 2012 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

TRANSPARENT = 'TRANSPARENT'
BLACK = 'BLACK'
BLUE = 'BLUE'
RED = 'RED'
MAGENTA = 'MAGENTA'
GREEN = 'GREEN'
CYAN = 'CYAN'
YELLOW = 'YELLOW'
WHITE = 'WHITE'
BRIGHT_BLUE = 'BRIGHT_BLUE'
BRIGHT_RED = 'BRIGHT_RED'
BRIGHT_MAGENTA = 'BRIGHT_MAGENTA'
BRIGHT_GREEN = 'BRIGHT_GREEN'
BRIGHT_CYAN = 'BRIGHT_CYAN'
BRIGHT_YELLOW = 'BRIGHT_YELLOW'
BRIGHT_WHITE = 'BRIGHT_WHITE'

# Image writer options
DEFAULT_FORMAT = 'DefaultFormat'
PNG_COMPRESSION_LEVEL = 'PNGCompressionLevel'
PNG_ALPHA = 'PNGAlpha'
PNG_ENABLE_ANIMATION = 'PNGEnableAnimation'
GIF_ENABLE_ANIMATION = 'GIFEnableAnimation'
GIF_TRANSPARENCY = 'GIFTransparency'
GIF_COMPRESSION = 'GIFCompression'

BITS4 = [[(n & m) // m for m in (8, 4, 2, 1)] for n in range(16)]
BITS8 = [[(n & m) // m for m in (128, 64, 32, 16, 8, 4, 2, 1)] for n in range(256)]
BIT_PAIRS = [((b & 192) >> 6, (b & 48) >> 4, (b & 12) >> 2, b & 3) for b in range(256)]
BD4_MAP0 = (
    ((0, 1),),
    ((0, 2), (1, 2)),
    ((1, 2), (0, 2)),
    ((1, 1),)
)
BD4_MAP1 = (
    ((0, 1),),
    ((0, 2), (1, 0), (3, 2)),
    ((3, 2), (2, 0), (0, 2)),
    ((3, 1),)
)
BD2_BITS = [BITS4[i] for i in (0, 1, 3, 7, 8, 12, 14, 15)]

# http://www.libpng.org/pub/png/spec/iso/index-object.html
# https://wiki.mozilla.org/APNG_Specification
PNG_SIGNATURE = (137, 80, 78, 71, 13, 10, 26, 10)
IHDR = (73, 72, 68, 82)
PLTE = (80, 76, 84, 69)
TRNS = (116, 82, 78, 83)
ACTL_CHUNK = (0, 0, 0, 8, 97, 99, 84, 76, 0, 0, 0, 2, 0, 0, 0, 0, 243, 141, 147, 112)
FCTL = (102, 99, 84, 76)
IDAT = (73, 68, 65, 84)
FDAT = (102, 100, 65, 84, 0, 0, 0, 2)
IEND_CHUNK = (0, 0, 0, 0, 73, 69, 78, 68, 174, 66, 96, 130)
CRC_MASK = 4294967295

# http://www.w3.org/Graphics/GIF/spec-gif89a.txt
GIF89A = (71, 73, 70, 56, 57, 97)
GIF_FRAME_DELAY = 32
AEB = (33, 255, 11, 78, 69, 84, 83, 67, 65, 80, 69, 50, 46, 48, 3, 1, 0, 0, 0)
GIF_TRAILER = 59

BINSTR = [''.join([str((m >> n) & 1) for n in range(11, -1, -1)]) for m in range(4096)]

class ImageWriter:
    def __init__(self, palette=None, options=None, build_images=True):
        self.options = self._get_default_options()
        if options:
            for k, v in options.items():
                if k == DEFAULT_FORMAT:
                    self.options[k] = v
                else:
                    try:
                        self.options[k] = int(v)
                    except ValueError:
                        pass
            self.options[PNG_ALPHA] &= 255
        self.build_images = build_images
        if build_images:
            self.trns = list(TRNS)
            full_palette = self._get_default_palette()
            if palette:
                full_palette.update(palette)
            self._create_colours(full_palette)
            self._create_attr_index()
            self._create_crc_table()
            self.write_image = self._write_image
            self._create_png_method_dict()
        else:
            self.write_image = self._null_write_image

    def _get_default_palette(self):
        return  {
            TRANSPARENT: (0, 254, 0),
            BLACK: (0, 0, 0),
            BLUE: (0, 0, 197),
            RED: (197, 0, 0),
            MAGENTA: (197, 0, 197),
            GREEN: (0, 198, 0),
            CYAN: (0, 198, 197),
            YELLOW: (197, 198, 0),
            WHITE: (205, 198, 205),
            BRIGHT_BLUE: (0, 0, 255),
            BRIGHT_RED: (255, 0, 0),
            BRIGHT_MAGENTA: (255, 0, 255),
            BRIGHT_GREEN: (0, 255, 0),
            BRIGHT_CYAN: (0, 255, 255),
            BRIGHT_YELLOW: (255, 255, 0),
            BRIGHT_WHITE: (255, 255, 255)
        }

    def _get_default_options(self):
        return {
            DEFAULT_FORMAT: 'png',
            PNG_COMPRESSION_LEVEL: 9,
            PNG_ENABLE_ANIMATION: 1,
            PNG_ALPHA: 255,
            GIF_ENABLE_ANIMATION: 1,
            GIF_TRANSPARENCY: 0,
            GIF_COMPRESSION: 1
        }

    def _create_colours(self, palette):
        colours = []
        colours.append(palette[TRANSPARENT])
        colours.append(palette[BLACK])
        colours.append(palette[BLUE])
        colours.append(palette[RED])
        colours.append(palette[MAGENTA])
        colours.append(palette[GREEN])
        colours.append(palette[CYAN])
        colours.append(palette[YELLOW])
        colours.append(palette[WHITE])
        colours.append(palette[BRIGHT_BLUE])
        colours.append(palette[BRIGHT_RED])
        colours.append(palette[BRIGHT_MAGENTA])
        colours.append(palette[BRIGHT_GREEN])
        colours.append(palette[BRIGHT_CYAN])
        colours.append(palette[BRIGHT_YELLOW])
        colours.append(palette[BRIGHT_WHITE])
        self.colours = colours

    def _create_attr_index(self):
        self.attr_index = {}
        for attr in range(128):
            if attr & 64:
                ink = 8 + (attr & 7)
                paper = 8 + (attr & 56) // 8
                if ink == 8:
                    ink = 1
                if paper == 8:
                    paper = 1
            else:
                ink = 1 + (attr & 7)
                paper = 1 + (attr & 56) // 8
            self.attr_index[attr] = (paper, ink)

    def _create_crc_table(self):
        self.crc_table = []
        for n in range(256):
            c = n
            for k in range(8):
                if c & 1:
                    c = 3988292384 ^ (c >> 1)
                else:
                    c = c >> 1
            self.crc_table.append(c)

    def _create_png_method_dict(self):
        # The PNG method dictionary is keyed on:
        #   bit_depth: 0 (1 colour), 1 (2 colours), 2 or 4
        #   full_size: 0 (cropped), 1 (full size)
        #   trans: 0 (no transparent colour), 1 (some UDGs have masks), or 2
        #          (all UDGs have masks)
        #   flash: 1 (invert paper/ink for flashing cells), 0 (don't)
        # and the values are 2-element tuples:
        #   (method1, method2)
        # where method1, if not None, returns the appropriate method to use to
        # build the image data (based on the number of UDGs, the number of
        # distinct attributes, and the scale); otherwise method2 is used to
        # build the image data

        # Create default values
        self.png_method_dict = {}
        default_methods = (None, self._build_image_data_bd_any)
        for bit_depth in (0, 1, 2, 4):
            bd_method_dict = self.png_method_dict[bit_depth] = {}
            for full_size in (0, 1):
                fs_method_dict = bd_method_dict[full_size] = {}
                for trans in (0, 1, 2):
                    ht_method_dict = fs_method_dict[trans] = {}
                    for flash in (0, 1):
                        ht_method_dict[flash] = default_methods

    def _to_bytes(self, num):
        return (num >> 24, (num >> 16) & 255, (num >> 8) & 255, num & 255)

    def _get_colours(self, udg_array, scale, mask, x, y, width, height, use_flash):
        # Find all the colours in a cropped image
        attrs = set()
        colours = set()
        has_trans = 0
        all_masked = 1
        flashing = False
        y_count = 0
        inc = 8 * scale
        min_x = width
        min_y = height
        max_x = max_y = 0
        rows = 0
        for row in udg_array:
            if y_count >= y + height:
                break
            if y_count + inc - 1 < y:
                y_count += inc
                continue
            for j in range(8):
                if y_count + scale - 1 < y:
                    y_count += scale
                    continue
                if y_count >= y:
                    num_lines = min(y + height - y_count, scale)
                else:
                    num_lines = y_count - y + scale
                x_count = 0
                cols = 0
                for udg in row:
                    if x_count >= x + width:
                        break
                    if x_count + inc - 1 < x:
                        x_count += inc
                        continue
                    attr = udg.attr & 127
                    attrs.add(attr)
                    paper, ink = self.attr_index[attr]
                    byte = udg.data[j]
                    if mask and udg.mask:
                        mask_byte = udg.mask[j]
                        byte &= mask_byte
                    else:
                        all_masked = 0
                        mask_byte = 0
                    if use_flash and udg.attr & 128:
                        colours.add(ink)
                        if ink != paper:
                            colours.add(paper)
                            flashing = True
                    for k in range(8):
                        if x_count >= x + width:
                            break
                        if x_count + scale - 1 < x:
                            x_count += scale
                            byte *= 2
                            mask_byte *= 2
                            continue
                        if x_count >= x:
                            num_bits = min(x + width - x_count, scale)
                        else:
                            num_bits = x_count - x + scale
                        if use_flash and udg.attr & 128 and ink != paper:
                            min_x = min((cols, min_x))
                            max_x = max((cols + num_bits, max_x))
                            min_y = min((rows, min_y))
                            max_y = max((rows + num_lines, max_y))
                        if byte & 128:
                            colours.add(ink)
                        elif mask_byte & 128:
                            has_trans = 1
                        else:
                            colours.add(paper)
                        byte *= 2
                        mask_byte *= 2
                        x_count += scale
                        cols += num_bits
                y_count += scale
                rows += num_lines

        if flashing:
            flash_rect = (min_x, min_y, max_x - min_x, max_y - min_y)
        else:
            flash_rect = None
        trans = 1 + all_masked if has_trans else 0
        return colours, attrs, trans, flash_rect

    def _get_all_colours(self, udg_array, mask, use_flash):
        # Find all the colours in an uncropped image
        attrs = set()
        colours = set()
        has_trans = 0
        all_masked = 1
        flashing = False
        min_x = len(udg_array[0])
        min_y = len(udg_array)
        max_x = max_y = 0

        y = 0
        for row in udg_array:
            x = 0
            for udg in row:
                attr = udg.attr
                attrs.add(attr & 127)
                paper, ink = self.attr_index[attr & 127]
                if mask and udg.mask:
                    data = [udg.data[i] & udg.mask[i] for i in range(8)]
                    has_non_trans = False
                    if any(data):
                        colours.add(ink)
                        has_non_trans = True
                    if any([b < 255 for b in udg.mask]):
                        colours.add(paper)
                        has_non_trans = True
                    if use_flash and attr & 128 and has_non_trans:
                        colours.add(ink)
                        if ink != paper:
                            min_x = min((x, min_x))
                            min_y = min((y, min_y))
                            max_x = max((x, max_x))
                            max_y = max((y, max_y))
                            colours.add(paper)
                            flashing = True
                    if any([data[i] | udg.mask[i] > data[i] for i in range(8)]):
                        has_trans = 1
                elif use_flash and attr & 128:
                    colours.add(ink)
                    if ink != paper:
                        min_x = min((x, min_x))
                        min_y = min((y, min_y))
                        max_x = max((x, max_x))
                        max_y = max((y, max_y))
                        colours.add(paper)
                        flashing = True
                    all_masked = 0
                else:
                    if any(udg.data):
                        colours.add(ink)
                    if any([b < 255 for b in udg.data]):
                        colours.add(paper)
                    all_masked = 0
                x += 1
            y += 1

        if flashing:
            flash_rect = (min_x, min_y, max_x - min_x + 1, max_y - min_y + 1)
        else:
            flash_rect = None
        trans = 1 + all_masked if has_trans else 0
        return colours, attrs, trans, flash_rect

    def _get_palette(self, udg_array, scale, mask, x, y, width, height, full_size, use_flash):
        if full_size:
            colours, attrs, trans, flash_rect = self._get_all_colours(udg_array, mask, use_flash)
        else:
            colours, attrs, trans, flash_rect = self._get_colours(udg_array, scale, mask, x, y, width, height, use_flash)

        colour_map = {}
        palette = []
        i = 0
        if trans:
            palette.extend(self.colours[0])
            i += 1
        for colour in colours:
            palette.extend(self.colours[colour])
            colour_map[colour] = i
            i += 1

        attr_map = {}
        for attr in attrs:
            paper, ink = self.attr_index[attr]
            attr_map[attr] = (colour_map.get(paper, 0), colour_map.get(ink, 0))

        return palette, attr_map, trans, flash_rect

    def _build_image_data(self, udg_array, scale, mask, x, y, width, height, use_flash):
        t_width = len(udg_array[0])
        t_height = len(udg_array)
        full_width = 8 * t_width * scale
        full_height = 8 * t_height * scale
        width = min(width or full_width, full_width - x)
        height = min(height or full_height, full_height - y)
        full_size = 1 if width == full_width and height == full_height else 0

        palette, attr_map, trans, flash_rect = self._get_palette(udg_array, scale, mask, x, y, width, height, full_size, use_flash)
        palette_size = len(palette) // 3
        if palette_size > 4:
            bit_depth = 4
        elif palette_size > 2:
            bit_depth = 2
        else:
            bit_depth = 1
        plte_chunk = list(PLTE)
        plte_chunk.extend(palette)

        frame2 = frame2_rect = None
        if flash_rect:
            if full_size:
                f2_tx, f2_ty, f2_tw, f2_th = flash_rect
                sf = 8 * scale
                frame2_rect = (f2_tx * sf, f2_ty * sf, f2_tw * sf, f2_th * sf)
            else:
                frame2_rect = flash_rect

        bd = 0 if palette_size == 1 else bit_depth
        method_dict = self.png_method_dict[bd][full_size][trans]

        # Frame 1
        f1_method, f1_build_method = method_dict[0]
        if f1_method:
            f1_build_method = f1_method(t_width * t_height, len(attr_map), scale)
        frame1 = f1_build_method(udg_array, scale, attr_map, trans, False, x, y, width, height, bit_depth)

        # Frame 2
        if frame2_rect:
            if full_size and (f2_tw < len(udg_array[0]) or f2_th < len(udg_array)):
                f2_udg_array = [udg_array[i][f2_tx:f2_tx + f2_tw] for i in range(f2_ty, f2_ty + f2_th)]
            else:
                f2_udg_array = udg_array
            f2_x, f2_y, f2_w, f2_h = frame2_rect
            f2_method, f2_build_method = method_dict[1]
            if f2_method:
                f2_build_method = f2_method(len(f2_udg_array[0]) * len(f2_udg_array), len(attr_map), scale)
            frame2 = f2_build_method(f2_udg_array, scale, attr_map, trans, True, x + f2_x, y + f2_y, f2_w, f2_h, bit_depth)

        return width, height, bit_depth, plte_chunk, trans, frame1, frame2, frame2_rect

    def _write_ihdr_chunk(self, img_file, width, height, bit_depth):
        data = list(IHDR) # chunk type
        data.extend(self._to_bytes(width)) # width
        data.extend(self._to_bytes(height)) # height
        data.extend((bit_depth, 3)) # bit depth and colour type
        data.extend((0, 0, 0)) # compression, filter and interlace methods
        self._write_chunk(img_file, data)

    def _write_fctl_chunk(self, img_file, frame_num, width, height, x_offset=0, y_offset=0):
        data = list(FCTL) # chunk type
        data.extend(self._to_bytes(frame_num)) # frame number
        data.extend(self._to_bytes(width)) # width
        data.extend(self._to_bytes(height)) # height
        data.extend(self._to_bytes(x_offset)) # x offset
        data.extend(self._to_bytes(y_offset)) # y offset
        data.extend((0, 8)) # delay numerator
        data.extend((0, 25)) # delay denominator
        data.append(0) # frame disposal operation
        data.append(0) # frame blending operation
        self._write_chunk(img_file, data)

    def _write_png(self, udg_array, img_file, scale=1, mask=False, x=0, y=0, width=None, height=None):
        use_flash = self.options[PNG_ENABLE_ANIMATION]
        width, height, bit_depth, plte_chunk, trans, frame1, frame2, frame2_rect = self._build_image_data(udg_array, scale, mask, x, y, width, height, use_flash)

        # PNG signature
        img_file.write(self.png_signature)

        # IHDR
        self._write_ihdr_chunk(img_file, width, height, bit_depth)

        # PLTE
        self._write_chunk(img_file, plte_chunk)

        # tRNS
        if trans and self.options[PNG_ALPHA] != 255:
            self._write_chunk(img_file, self.trns + [self.options[PNG_ALPHA]])

        # acTL and fcTL
        if frame2:
            img_file.write(self.actl_chunk)
            self._write_fctl_chunk(img_file, 0, width, height)

        # IDAT
        self._write_img_data_chunk(img_file, self.idat + frame1)

        # fcTL and fdAT
        if frame2:
            f2_x, f2_y, f2_w, f2_h = frame2_rect
            self._write_fctl_chunk(img_file, 1, f2_w, f2_h, f2_x, f2_y)
            self._write_img_data_chunk(img_file, self.fdat + frame2)

        # IEND
        img_file.write(self.iend_chunk)

    def _get_pixels(self, udg_array, scale, trans, x, y, width, height, flash, attr_map):
        pixels = []
        y_count = 0
        inc = 8 * scale
        for row in udg_array:
            if y_count >= y + height:
                break
            if y_count + inc <= y:
                y_count += inc
                continue
            for k in range(8):
                if y_count + scale <= y:
                    y_count += scale
                    continue
                if y_count >= y:
                    num_lines = min(y + height - y_count, scale)
                else:
                    num_lines = y_count - y + scale
                y_count += scale
                x_count = 0
                pixel_row = []
                for udg in row:
                    if x_count >= x + width:
                        break
                    if x_count + inc <= x:
                        x_count += inc
                        continue
                    paper, ink = attr_map[udg.attr & 127]
                    if flash and udg.attr & 128:
                        paper, ink = ink, paper
                    byte = udg.data[k]
                    if trans and udg.mask:
                        mask_byte = udg.mask[k]
                        byte &= mask_byte
                    else:
                        mask_byte = 0
                    for j in range(8):
                        if x_count >= x + width:
                            break
                        if x_count + scale <= x:
                            x_count += scale
                            byte *= 2
                            mask_byte *= 2
                            continue
                        if x_count >= x:
                            num_bits = min(x + width - x_count, scale)
                        else:
                            num_bits = x_count - x + scale
                        x_count += scale
                        if byte & 128:
                            pixel_row.extend((ink,) * num_bits)
                        elif mask_byte & 128 == 0:
                            pixel_row.extend((paper,) * num_bits)
                        else:
                            pixel_row.extend((0,) * num_bits)
                        byte *= 2
                        mask_byte *= 2
                for n in range(num_lines):
                    pixels += pixel_row
        return pixels

    def _get_all_pixels(self, udg_array, scale, trans, flash, attr_map):
        attr_index = {}
        for attr, (paper, ink) in attr_map.items():
            attr_index[attr] = ((paper,) * scale, (ink,) * scale)

        pixels = []
        t = (0,) * scale
        for row in udg_array:
            for k in range(8):
                pixel_row = []
                for udg in row:
                    p, i = attr_index[udg.attr & 127]
                    if flash and udg.attr & 128:
                        p, i = i, p
                    byte = udg.data[k]
                    if trans and udg.mask:
                        mask_byte = udg.mask[k]
                        byte &= mask_byte
                    else:
                        mask_byte = 0
                    for j in range(8):
                        if byte & 128:
                            pixel_row.extend(i)
                        elif mask_byte & 128 == 0:
                            pixel_row.extend(p)
                        else:
                            pixel_row.extend(t)
                        byte *= 2
                        mask_byte *= 2
                for n in range(scale):
                    pixels += pixel_row
        return pixels

    def _get_pixels_as_str(self, udg_array, scale, trans, x, y, width, height, flash, attr_map):
        pixels = []
        y_count = 0
        inc = 8 * scale
        for row in udg_array:
            if y_count >= y + height:
                break
            if y_count + inc <= y:
                y_count += inc
                continue
            for k in range(8):
                if y_count + scale <= y:
                    y_count += scale
                    continue
                if y_count >= y:
                    num_lines = min(y + height - y_count, scale)
                else:
                    num_lines = y_count - y + scale
                y_count += scale
                x_count = 0
                pixel_row = []
                for udg in row:
                    if x_count >= x + width:
                        break
                    if x_count + inc <= x:
                        x_count += inc
                        continue
                    paper, ink = attr_map[udg.attr & 127]
                    if flash and udg.attr & 128:
                        paper, ink = ink, paper
                    byte = udg.data[k]
                    if trans and udg.mask:
                        mask_byte = udg.mask[k]
                        byte &= mask_byte
                    else:
                        mask_byte = 0
                    for j in range(8):
                        if x_count >= x + width:
                            break
                        if x_count + scale <= x:
                            x_count += scale
                            byte *= 2
                            mask_byte *= 2
                            continue
                        if x_count >= x:
                            num_bits = min(x + width - x_count, scale)
                        else:
                            num_bits = x_count - x + scale
                        x_count += scale
                        if byte & 128:
                            pixel_row.append(chr(ink) * num_bits)
                        elif mask_byte & 128 == 0:
                            pixel_row.append(chr(paper) * num_bits)
                        else:
                            pixel_row.append(chr(0) * num_bits)
                        byte *= 2
                        mask_byte *= 2
                for n in range(num_lines):
                    pixels += pixel_row
        return ''.join(pixels)

    def _get_all_pixels_as_str(self, udg_array, scale, trans, flash, attr_map):
        attr_index = {}
        for attr, (paper, ink) in attr_map.items():
            attr_index[attr] = (chr(paper) * scale, chr(ink) * scale)

        pixels = []
        t = chr(0) * scale
        for row in udg_array:
            for k in range(8):
                pixel_row = []
                for udg in row:
                    p, i = attr_index[udg.attr & 127]
                    if flash and udg.attr & 128:
                        p, i = i, p
                    byte = udg.data[k]
                    if trans and udg.mask:
                        mask_byte = udg.mask[k]
                        byte &= mask_byte
                    else:
                        mask_byte = 0
                    for j in range(8):
                        if byte & 128:
                            pixel_row.append(i)
                        elif mask_byte & 128 == 0:
                            pixel_row.append(p)
                        else:
                            pixel_row.append(t)
                        byte *= 2
                        mask_byte *= 2
                for n in range(scale):
                    pixels += pixel_row
        return ''.join(pixels)

    def _compress(self, pixels, min_code_size):
        # Initialise the dictionary
        init_d = {}
        for n in range(1 << min_code_size):
            init_d[chr(n)] = n

        # Add initial (dummy) STOP and CLEAR codes
        clear_code = 1 << min_code_size
        stop_code = clear_code + 1
        init_d[clear_code] = 0
        init_d[stop_code] = 0

        d = init_d.copy()
        code_size = min_code_size + 1
        output = []
        bit_buf = BINSTR[clear_code][-code_size:]
        i = 0
        while 1:
            # Check for max dictionary length
            if len(d) == 4096:
                # Output a CLEAR code
                bit_buf = BINSTR[clear_code] + bit_buf
                # Initialise the dictionary and reset the code size
                d = init_d.copy()
                code_size = min_code_size + 1

            # Find the next substring not yet in the dictionary
            j = 1
            while i + j <= len(pixels):
                new_substr = pixels[i:i + j]
                if new_substr in d:
                    substr = new_substr
                    j += 1
                else:
                    break
            bit_buf = BINSTR[d[substr]][-code_size:] + bit_buf

            # Flush any full bytes in the bit buffer to the output
            while len(bit_buf) > 7:
                output.append(int(bit_buf[-8:], 2))
                bit_buf = bit_buf[:-8]

            if new_substr not in d:
                if len(d) == 1 << code_size:
                    code_size += 1
                # Add the code for the new substring
                d[new_substr] = len(d)
                i += len(substr)
            else:
                break

        # Output the STOP code
        bit_buf = BINSTR[stop_code][-code_size:] + bit_buf

        # Flush any remaining bits from the buffer
        while bit_buf:
            output.append(int(bit_buf[-8:], 2))
            bit_buf = bit_buf[:-8]

        return output

    def _write_gct(self, img_file, palette):
        data = []
        gct_flag = 1 # Global Colour Table (GCT) follows
        colour_res = 7 # number of bits per RGB channel = colour_res + 1 = 8
        sort_flag = 0 # GCT is not sorted
        palette_size = len(palette) // 3
        if palette_size > 8:
            gct_size = 3
            num_entries = 16
            min_code_size = 4
        elif palette_size > 4:
            gct_size = 2
            num_entries = 8
            min_code_size = 3
        elif palette_size > 2:
            gct_size = 1
            num_entries = 4
            min_code_size = 2
        else:
            gct_size = 0
            num_entries = 2
            min_code_size = 2
        data.append(128 * gct_flag + 16 * colour_res + 8 * sort_flag + gct_size)
        data.append(0) # background colour index (redundant)
        data.append(0) # pixel aspect ratio (none)
        data.extend(palette)
        for i in range(num_entries - palette_size):
            data.extend((0, 0, 0))
        self._write_block(img_file, data)
        return min_code_size

    def _write_gce(self, img_file, delay, transparent):
        data = []
        data.extend((33, 249)) # Graphic Control Extension
        data.append(4) # 4 bytes follow
        data.append(1 if transparent else 0) # transparency
        data.extend((delay % 256, delay // 256)) # delay before next frame
        data.append(0) # transparent colour
        data.append(0) # end of GCE block
        self._write_block(img_file, data)

    def _write_image_descriptor(self, img_file, width, height, x_offset=0, y_offset=0):
        data = []
        data.append(44) # Image separator
        data.extend((x_offset % 256, x_offset // 256)) # image left position
        data.extend((y_offset % 256, y_offset // 256)) # image top position
        data.extend((width % 256, width // 256)) # image width
        data.extend((height % 256, height // 256)) # image height
        data.append(0) # no local colour table
        self._write_block(img_file, data)

    def _write_gif_image_data(self, img_file, udg_array, scale, trans, x, y, width, height, full_size, min_code_size, attr_map, flash=False):
        if self.options[GIF_COMPRESSION]:
            data = [min_code_size]
            if full_size:
                pixels = self._get_all_pixels_as_str(udg_array, scale, trans, flash, attr_map)
            else:
                pixels = self._get_pixels_as_str(udg_array, scale, trans, x, y, width, height, flash, attr_map)
            pixels_lzw = self._compress(pixels, min_code_size)
            p_count = len(pixels_lzw)
            m = 0
            while p_count:
                b_count = min((p_count, 255))
                p_count -= b_count
                data.append(b_count) # This many bytes follow
                data += pixels_lzw[m:m + b_count]
                m += b_count
        else:
            data = [7] # Minimum code size for uncompressed LZW data
            if full_size:
                pixels = self._get_all_pixels(udg_array, scale, trans, flash, attr_map)
            else:
                pixels = self._get_pixels(udg_array, scale, trans, x, y, width, height, flash, attr_map)
            pixels.append(129) # 8-bit STOP code
            p_count = len(pixels)
            m = 0
            while p_count:
                # 127 is the maximum distance between CLEAR codes in an
                # uncompressed LZW stream
                b_count = min((p_count, 126))
                p_count -= b_count
                data.append(b_count + 1) # This many bytes follow
                data.append(128) # 8-bit CLEAR code
                data += pixels[m:m + b_count]
                m += b_count
        data.append(0) # End of image data
        self._write_block(img_file, data)

    def _write_gif(self, udg_array, img_file, scale, mask, x, y, width, height):
        use_flash = self.options[GIF_ENABLE_ANIMATION]
        full_width = 8 * len(udg_array[0]) * scale
        full_height = 8 * len(udg_array) * scale
        width = min(width or full_width, full_width - x)
        height = min(height or full_height, full_height - y)
        full_size = width == full_width and height == full_height
        palette, attr_map, trans, flash_rect = self._get_palette(udg_array, scale, mask, x, y, width, height, full_size, use_flash)
        transparent = self.options[GIF_TRANSPARENCY] and trans

        # Header
        img_file.write(self.gif_header)

        # Logical screen descriptor
        data = []
        data.extend((width % 256, width // 256)) # logical screen width
        data.extend((height % 256, height // 256)) # logical screen height
        self._write_block(img_file, data)

        # Global Colour Table
        min_code_size = self._write_gct(img_file, palette)

        if flash_rect:
            img_file.write(self.aeb)
            self._write_gce(img_file, GIF_FRAME_DELAY, transparent)
        elif transparent:
            self._write_gce(img_file, 0, transparent)

        # Frame 1
        self._write_image_descriptor(img_file, width, height)
        self._write_gif_image_data(img_file, udg_array, scale, trans, x, y, width, height, full_size, min_code_size, attr_map)

        if flash_rect:
            # Frame 2
            f2_x, f2_y, f2_w, f2_h = flash_rect
            f2_udg_array = udg_array
            if full_size:
                if f2_w < width or f2_h < height:
                    f2_udg_array = [udg_array[i][f2_x:f2_x + f2_w] for i in range(f2_y, f2_y + f2_h)]
                f2_x *= 8 * scale
                f2_y *= 8 * scale
                f2_w *= 8 * scale
                f2_h *= 8 * scale
            self._write_gce(img_file, GIF_FRAME_DELAY, transparent)
            self._write_image_descriptor(img_file, f2_w, f2_h, f2_x, f2_y)
            self._write_gif_image_data(img_file, f2_udg_array, scale, trans, x + f2_x, y + f2_y, f2_w, f2_h, full_size, min_code_size, attr_map, True)

        # GIF trailer
        img_file.write(self.gif_trailer)

    def _write_image(self, udg_array, img_file, format, scale=1, mask=False, x=0, y=0, width=None, height=None):
        if format == 'png':
            self._write_png(udg_array, img_file, scale, mask, x, y, width, height)
        elif format == 'gif':
            self._write_gif(udg_array, img_file, scale, mask, x, y, width, height)

    def _null_write_image(self, udgs, img_file, format, scale=1, mask=False, x=0, y=0, width=None, height=None):
        return

    def get_default_format(self):
        return self.options[DEFAULT_FORMAT]
