# -*- coding: utf-8 -*-

# Copyright 2012 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import zlib
from .writer import ImageWriter, PNG_SIGNATURE, ACTL_CHUNK, IDAT, FDAT, IEND_CHUNK, GIF89A, AEB, GIF_TRAILER, PNG_COMPRESSION_LEVEL, CRC_MASK, BIT_PAIRS, BITS4, BITS8, BD2_BITS, BD4_MAP0, BD4_MAP1

# Characters for all possible bit patterns at even scales at bit depth 2
BD2_CHARS2 = (
    chr(0),   # 00000000 (0000)
    chr(5),   # 00000101 (0011)
    chr(10),  # 00001010 (0022)
    chr(15),  # 00001111 (0033)
    chr(80),  # 01010000 (1100)
    chr(85),  # 01010101 (1111)
    chr(90),  # 01011010 (1122)
    chr(95),  # 01011111 (1133)
    chr(160), # 10100000 (2200)
    chr(165), # 10100101 (2211)
    chr(170), # 10101010 (2222)
    chr(175), # 10101111 (2233)
    chr(240), # 11110000 (3300)
    chr(245), # 11110101 (3311)
    chr(250), # 11111010 (3322)
    chr(255)  # 11111111 (3333)
)

class ImageWriter2(ImageWriter):
    def __init__(self, palette=None, options=None, build_images=True):
        ImageWriter.__init__(self, palette, options, build_images)
        if build_images:
            self.png_signature = ''.join([chr(b) for b in PNG_SIGNATURE])
            self.actl_chunk = ''.join([chr(b) for b in ACTL_CHUNK])
            self.idat = ''.join([chr(b) for b in IDAT])
            self.fdat = ''.join([chr(b) for b in FDAT])
            self.iend_chunk = ''.join([chr(b) for b in IEND_CHUNK])
            self.gif_header = ''.join([chr(b) for b in GIF89A])
            self.aeb = ''.join([chr(b) for b in AEB])
            self.gif_trailer = chr(GIF_TRAILER)

    def _create_png_method_dict(self):
        ImageWriter._create_png_method_dict(self)

        # Bit depth 0 (1 colour)
        bd0_methods = (None, self._build_image_data_bd1_blank)
        bd0_fs_method_dict = self.png_method_dict[0][1]
        bd0_fs_method_dict[0][0] = bd0_methods
        bd0_fs_method_dict[0][1] = bd0_methods
        bd0_fs_method_dict[1][0] = bd0_methods
        bd0_fs_method_dict[1][1] = bd0_methods
        bd0_fs_method_dict[2][0] = bd0_methods
        bd0_fs_method_dict[2][1] = bd0_methods

        # Bit depth 1 (2 colours)
        bd1_methods = (self._bd1_method, None)
        bd1_fs_method_dict = self.png_method_dict[1][1]
        bd1_fs_method_dict[0][0] = (self._bd1_nt_nf_method, None)
        bd1_fs_method_dict[0][1] = bd1_methods
        bd1_fs_method_dict[1][0] = bd1_methods
        bd1_fs_method_dict[1][1] = bd1_methods
        bd1_fs_method_dict[2][0] = bd1_methods
        bd1_fs_method_dict[2][1] = bd1_methods

        # Bit depth 2
        bd2_methods = (None, self._build_image_data_bd2)
        bd2_fs_method_dict = self.png_method_dict[2][1]
        bd2_fs_method_dict[0][0] = (None, self._build_image_data_bd2_nt_nf)
        bd2_fs_method_dict[0][1] = bd2_methods
        bd2_fs_method_dict[1][0] = bd2_methods
        bd2_fs_method_dict[1][1] = bd2_methods
        bd2_fs_method_dict[2][0] = (self._bd2_at_nf_method, None)
        bd2_fs_method_dict[2][1] = bd2_methods

        # Bit depth 4
        bd4_methods = (None, self._build_image_data_bd4)
        bd4_fs_method_dict = self.png_method_dict[4][1]
        bd4_fs_method_dict[0][0] = (self._bd4_nt_nf_method, None)
        bd4_fs_method_dict[0][1] = bd4_methods
        bd4_fs_method_dict[1][0] = bd4_methods
        bd4_fs_method_dict[1][1] = bd4_methods
        bd4_fs_method_dict[2][0] = bd4_methods
        bd4_fs_method_dict[2][1] = bd4_methods

    def _bd4_nt_nf_method(self, num_udgs, num_attrs, scale):
        if scale == 2:
            min_index = 18
        elif scale == 1:
            min_index = 11
        elif scale & 1:
            min_index = 55
        else:
            min_index = 47
        if num_udgs / num_attrs >= min_index:
            return self._build_image_data_bd4_nt_nf
        return self._build_image_data_bd4

    def _bd2_at_nf_method(self, num_udgs, num_attrs, scale):
        if scale == 2 or scale & 3 == 0:
            return self._build_image_data_bd2_at_nf
        return self._build_image_data_bd2

    def _bd1_method(self, num_udgs, num_attrs, scale):
        if num_udgs >= 5:
            return self._build_image_data_bd1
        if num_attrs == 1 and num_udgs >= 3:
            return self._build_image_data_bd1
        if num_attrs == 2 and num_udgs >= 4:
            return self._build_image_data_bd1
        return self._build_image_data_bd_any

    def _bd1_nt_nf_method(self, num_udgs, num_attrs, scale):
        if num_udgs == 1 and scale in (1, 2, 4, 8):
            return self._build_image_data_bd1_nt_nf_1udg
        return self._bd1_method(num_udgs, num_attrs, scale)

    def _write_block(self, img_file, block_data):
        img_file.write(''.join([chr(b) for b in block_data]))

    def _to_string(self, num):
        return ''.join([chr(b) for b in self._to_bytes(num)])

    def _get_crc(self, bytes):
        crc = CRC_MASK
        for b in bytes:
            crc = self.crc_table[(crc ^ b) & 255] ^ (crc >> 8)
        return self._to_string(crc ^ CRC_MASK)

    def _get_string_crc(self, chars):
        crc = CRC_MASK
        for c in chars:
            crc = self.crc_table[(crc ^ ord(c)) & 255] ^ (crc >> 8)
        return self._to_string(crc ^ CRC_MASK)

    def _write_chunk(self, img_file, chunk_data):
        img_file.write(self._to_string(len(chunk_data) - 4))
        img_file.write(''.join([chr(b) for b in chunk_data]))
        img_file.write(self._get_crc(chunk_data))

    def _write_img_data_chunk(self, img_file, img_data):
        img_file.write(self._to_string(len(img_data) - 4))
        img_file.write(img_data)
        img_file.write(self._get_string_crc(img_data))

    def _build_image_data_bd_any(self, udg_array, scale, attr_map, trans, flash, x, y, width, height, bit_depth):
        # Build image data at any bit depth using a generic method
        if bit_depth == 1:
            max_length = 2 + width // 8 if width & 7 else 1 + width // 8
        elif bit_depth == 2:
            max_length = 2 + width // 4 if width & 3 else 1 + width // 4
        elif bit_depth == 4:
            max_length = 2 + width // 2 if width & 1 else 1 + width // 2

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data = []
        y_count = 0
        inc = 8 * scale
        for row in udg_array:
            if y_count >= y + height:
                break
            if y_count + inc <= y:
                y_count += inc
                continue
            for i in range(8):
                if y_count + scale <= y:
                    y_count += scale
                    continue
                if y_count >= y:
                    num_lines = min(y + height - y_count, scale)
                else:
                    num_lines = y_count - y + scale
                y_count += scale
                p = []
                x_count = 0
                for udg in row:
                    if x_count >= x + width:
                        break
                    if x_count + inc <= x:
                        x_count += inc
                        continue
                    byte = udg.data[i]
                    attr = udg.attr
                    paper, ink = attr_map[attr & 127]
                    if flash and attr & 128:
                        paper, ink = ink, paper
                    if trans and udg.mask:
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                    else:
                        mask_byte = 0
                    for b in range(8):
                        if x_count >= x + width:
                            break
                        if x_count + scale <= x:
                            x_count += scale
                            byte *= 2
                            mask_byte *= 2
                            continue
                        if x_count >= x:
                            num_bits = min(x + width - x_count, scale)
                        else:
                            num_bits = x_count - x + scale
                        x_count += scale
                        if byte & 128:
                            p.extend((ink,) * num_bits)
                        elif mask_byte & 128 == 0:
                            p.extend((paper,) * num_bits)
                        else:
                            p.extend((0,) * num_bits)
                        byte *= 2
                        mask_byte *= 2
                scanline = [chr(0)]
                if bit_depth == 1:
                    p.extend((0,) * (8 - len(p) & 7))
                    scanline.extend([chr(p[j] * 128 + p[j + 1] * 64 + p[j + 2] * 32 + p[j + 3] * 16 + p[j + 4] * 8 + p[j + 5] * 4 + p[j + 6] * 2 + p[j + 7]) for j in range(0, len(p), 8)])
                elif bit_depth == 2:
                    p.extend((0,) * (4 - len(p) & 3))
                    scanline.extend([chr(p[j] * 64 + p[j + 1] * 16 + p[j + 2] * 4 + p[j + 3]) for j in range(0, len(p), 4)])
                elif bit_depth == 4:
                    p.extend((0,) * (2 - len(p) & 1))
                    scanline.extend([chr(p[j] * 16 + p[j + 1]) for j in range(0, len(p), 2)])
                scanline_str = ''.join(scanline)
                img_data.append(compressor.compress(scanline_str * num_lines))
        img_data.append(compressor.flush())
        return ''.join(img_data)

    def _build_image_data_bd4_nt_nf(self, udg_array, scale, attr_map, trans=0, flash=0, x=0, y=0, width=0, height=0, bit_depth=0):
        attrs = {}
        if scale == 1:
            for attr, (p, i) in attr_map.items():
                chars = (chr(p * 17), chr(p * 16 + i), chr(i * 16 + p), chr(i * 17))
                attrs[attr] = [''.join([chars[i] for i in BIT_PAIRS[j]]) for j in range(256)]
        elif scale == 2:
            for attr, (p, i) in attr_map.items():
                chars = (chr(p * 17), chr(i * 17))
                attrs[attr] = [''.join([chars[i] for i in BITS8[j]]) for j in range(256)]
        elif scale & 1:
            for attr, (p, i) in attr_map.items():
                chars = (chr(p * 17), chr(p * 16 + i), chr(i * 16 + p), chr(i * 17))
                attrs[attr] = [''.join([''.join([chars[n] * (scale // s if s else 1) for n, s in BD4_MAP1[i]]) for i in BIT_PAIRS[j]]) for j in range(256)]
        else:
            for attr, (p, i) in attr_map.items():
                chars = (chr(p * 17), chr(i * 17))
                attrs[attr] = [''.join([''.join([chars[n] * (scale // s) for n, s in BD4_MAP0[i]]) for i in BIT_PAIRS[j]]) for j in range(256)]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data_blocks = []
        for row in udg_array:
            for i in range(8):
                scanline = chr(0) + ''.join([attrs[udg.attr & 127][udg.data[i]] for udg in row])
                img_data_blocks.append(compressor.compress(scanline * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd4(self, udg_array, scale, attr_map, trans, flash, x=0, y=0, width=0, height=0, bit_depth=0):
        # Build image data with bit depth 4
        attr_dict = {}
        attr_dict_t = {}
        for attr, (p, i) in attr_map.items():
            attr_dict[attr] = (chr(p * 17), chr(p * 16 + i), chr(i * 16 + p), chr(i * 17))
            if trans:
                attr_dict_t[attr] = (chr(0), chr(p), chr(p * 16), chr(i), chr(i * 16))

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data_blocks = []
        scale_odd = scale & 1
        h_scale = scale // 2
        for row in udg_array:
            for i in range(8):
                scanline = [chr(0)]
                for udg in row:
                    attr = udg.attr
                    pp, pi, ip, ii = attr_dict[attr & 127]
                    if flash and attr & 128:
                        pp, pi, ip, ii = ii, ip, pi, pp
                    pph = pp * h_scale
                    iih = ii * h_scale
                    byte = udg.data[i]
                    if trans and udg.mask:
                        # Apply mask
                        tt, tp, pt, ti, it = attr_dict_t[attr & 127]
                        if flash and attr & 128:
                            tp, pt, ti, it = ti, it, tp, pt
                        tth = tt * h_scale
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        for b in range(4):
                            bits = byte & 192
                            mask_bits = mask_byte & 192
                            if bits == 192:
                                # ink + ink
                                scanline.append(ii * scale)
                            elif bits == 128:
                                scanline.append(iih)
                                if mask_bits & 64:
                                    # ink + transparent
                                    if scale_odd:
                                        scanline.append(it)
                                    scanline.append(tth)
                                else:
                                    # ink + paper
                                    if scale_odd:
                                        scanline.append(ip)
                                    scanline.append(pph)
                            elif bits == 64:
                                if mask_bits & 128:
                                    # transparent + ink
                                    scanline.append(tth)
                                    if scale_odd:
                                        scanline.append(ti)
                                else:
                                    # paper + ink
                                    scanline.append(pph)
                                    if scale_odd:
                                        scanline.append(pi)
                                scanline.append(iih)
                            else:
                                if mask_bits == 192:
                                    # transparent + transparent
                                    scanline.append(tt * scale)
                                elif mask_bits == 128:
                                    # transparent + paper
                                    scanline.append(tth)
                                    if scale_odd:
                                        scanline.append(tp)
                                    scanline.append(pph)
                                elif mask_bits == 64:
                                    # paper + transparent
                                    scanline.append(pph)
                                    if scale_odd:
                                        scanline.append(pt)
                                    scanline.append(tth)
                                else:
                                    # paper + paper
                                    scanline.append(pp * scale)
                            byte *= 4
                            mask_byte *= 4
                    else:
                        # No mask
                        if byte == 0:
                            scanline.append(pp * 4 * scale)
                        elif byte == 255:
                            scanline.append(ii * 4 * scale)
                        else:
                            for b in range(4):
                                bits = byte & 192
                                if bits == 0:
                                    # paper + paper
                                    scanline.append(pp * scale)
                                elif bits == 192:
                                    # ink + ink
                                    scanline.append(ii * scale)
                                elif bits == 128:
                                    # ink + paper
                                    scanline.append(iih)
                                    if scale_odd:
                                        scanline.append(ip)
                                    scanline.append(pph)
                                else:
                                    # paper + ink
                                    scanline.append(pph)
                                    if scale_odd:
                                        scanline.append(pi)
                                    scanline.append(iih)
                                byte *= 4
                img_data_blocks.append(compressor.compress(''.join(scanline) * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd2_nt_nf(self, udg_array, scale, attr_map, trans=0, flash=0, x=0, y=0, width=0, height=0, bit_depth=0):
        # Build image data with bit depth 2 for an image with no transparency
        scale_m = scale & 3
        q = scale // 4
        attrs = {}
        if scale_m == 2:
            for attr, (p, i) in attr_map.items():
                c = (chr(p * 85), chr(p * 80 + i * 5), chr(i * 80 + p * 5), chr(i * 85))
                attrs[attr] = [c[b3 * 3] * q + c[b3 * 2 + b2] + c[b2 * 3] * q + c[b1 * 3] * q + c[b1 * 2 + b0] + c[b0 * 3] * q for b3, b2, b1, b0 in BITS4]
        elif scale_m == 0:
            for attr, (p, i) in attr_map.items():
                c = (chr(p * 85), chr(i * 85))
                attrs[attr] = [c[b3] * q + c[b2] * q + c[b1] * q + c[b0] * q for b3, b2, b1, b0 in BITS4]
        elif scale_m == 1:
            if scale == 1:
                for attr, t in attr_map.items():
                    attrs[attr] = [chr(t[b3] * 64 + t[b2] * 16 + t[b1] * 4 + t[b0]) for b3, b2, b1, b0 in BITS4]
            else:
                for attr, t in attr_map.items():
                    c = [chr(t[b3] * 64 + t[b2] * 16 + t[b1] * 4 + t[b0]) for b3, b2, b1, b0 in BD2_BITS]
                    attrs[attr] = [c[b3 * 7] * q + c[b3 * 4 + b2 * 3] + c[b2 * 7] * (q - 1) + c[b2 * 5 + b1 * 2] + c[b1 * 7] * (q - 1) + c[b1 * 6 + b0] + c[b0 * 7] * q for b3, b2, b1, b0 in BITS4]
        else:
            for attr, t in attr_map.items():
                c = [chr(t[b3] * 64 + t[b2] * 16 + t[b1] * 4 + t[b0]) for b3, b2, b1, b0 in BD2_BITS]
                attrs[attr] = [c[b3 * 7] * q + c[b3 * 6 + b2] + c[b2 * 7] * q + c[b2 * 5 + b1 * 2] + c[b1 * 7] * q + c[b1 * 4 + b0 * 3] + c[b0 * 7] * q for b3, b2, b1, b0 in BITS4]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data_blocks = []
        for row in udg_array:
            for i in range(8):
                scanline = chr(0)
                for udg in row:
                    attr_chars = attrs[udg.attr & 127]
                    byte = udg.data[i]
                    scanline += attr_chars[(byte & 240) // 16] + attr_chars[byte & 15]
                img_data_blocks.append(compressor.compress(scanline * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd2_at_nf(self, udg_array, scale, attr_map, trans=2, flash=0, x=0, y=0, width=0, height=0, bit_depth=0):
        # Build image data with bit depth 2 for a masked image when scale & 3 == 0 or scale == 2
        q = scale // 4
        attrs = {}
        if scale & 3 == 0:
            chars = (chr(0) * q, chr(85) * q, chr(170) * q, chr(255) * q) # 00000000, 01010101, 10101010, 11111111
            for attr, (p, i) in attr_map.items():
                c = (chars[p], chars[0], chars[p], chars[i]) # 00 (paper), 01 (transparent), 10 (paper), 11 (ink)
                attrs[attr] = [c[b3 * 2 + m3] + c[b2 * 2 + m2] + c[b1 * 2 + m1] + c[b0 * 2 + m0] for b3, b2, b1, b0, m3, m2, m1, m0 in BITS8]
        else:
            # scale == 2
            for attr, (paper, ink) in attr_map.items():
                p = (paper, 0, paper, ink) # 00, 01, 10, 11
                attrs[attr] = [BD2_CHARS2[p[b3 * 2 + m3] * 4 + p[b2 * 2 + m2]] + BD2_CHARS2[p[b1 * 2 + m1] * 4 + p[b0 * 2 + m0]] for b3, b2, b1, b0, m3, m2, m1, m0 in BITS8]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        img_data_blocks = []
        for row in udg_array:
            for i in range(8):
                scanline = chr(0)
                for udg in row:
                    attr_chars = attrs[udg.attr & 127]
                    byte = udg.data[i]
                    mask_byte = udg.mask[i]
                    scanline += attr_chars[(byte & 240) + (mask_byte & 240) // 16] + attr_chars[(byte & 15) * 16 + (mask_byte & 15)]
                img_data_blocks.append(compressor.compress(scanline * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd2(self, udg_array, scale, attr_map, trans, flash, x=0, y=0, width=0, height=0, bit_depth=0):
        # Build image data with bit depth 2
        attr_dict = {}
        for attr, q in attr_map.items():
            attr_dict[attr] = [chr(q[b3] * 64 + q[b2] * 16 + q[b1] * 4 + q[b0]) for b3, b2, b1, b0 in BITS4]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        scale_m = scale & 3
        q_scale = scale // 4
        img_data_blocks = []
        for row in udg_array:
            for i in range(8):
                scanline = [chr(0)]
                for udg in row:
                    attr = udg.attr
                    byte = udg.data[i]
                    attrs = attr_dict[attr & 127]
                    if trans and udg.mask:
                        # Apply mask
                        p = []
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        paper, ink = attr_map[attr & 127]
                        if flash and attr & 128:
                            paper, ink = ink, paper
                        for b in range(8):
                            if byte & 128:
                                p.extend((ink,) * scale)
                            elif mask_byte & 128:
                                p.extend((0,) * scale)
                            else:
                                p.extend((paper,) * scale)
                            byte *= 2
                            mask_byte *= 2
                        scanline.extend([chr(p[j] * 64 + p[j + 1] * 16 + p[j + 2] * 4 + p[j + 3]) for j in range(0, len(p), 4)])
                    else:
                        # No mask
                        if flash and attr & 128:
                            byte ^= 255
                        for bits in ((byte & 240) // 16, byte & 15):
                            b3, b2, b1, b0 = BITS4[bits]
                            if scale_m == 0:
                                scanline.append(attrs[b3 * 15] * q_scale)
                                scanline.append(attrs[b2 * 15] * q_scale)
                                scanline.append(attrs[b1 * 15] * q_scale)
                                scanline.append(attrs[b0 * 15] * q_scale)
                            elif scale_m == 1:
                                if scale == 1:
                                    scanline.append(attrs[bits])
                                    continue
                                scanline.append(attrs[b3 * 15] * q_scale)
                                scanline.append(attrs[b3 * 8 + b2 * 7])
                                scanline.append(attrs[b2 * 15] * (q_scale - 1))
                                scanline.append(attrs[b2 * 12 + b1 * 3])
                                scanline.append(attrs[b1 * 15] * (q_scale - 1))
                                scanline.append(attrs[b1 * 14 + b0])
                                scanline.append(attrs[b0 * 15] * q_scale)
                            elif scale_m == 2:
                                scanline.append(attrs[b3 * 15] * q_scale)
                                scanline.append(attrs[b3 * 12 + b2 * 3])
                                scanline.append(attrs[b2 * 15] * q_scale)
                                scanline.append(attrs[b1 * 15] * q_scale)
                                scanline.append(attrs[b1 * 12 + b0 * 3])
                                scanline.append(attrs[b0 * 15] * q_scale)
                            else:
                                scanline.append(attrs[b3 * 15] * q_scale)
                                scanline.append(attrs[b3 * 14 + b2])
                                scanline.append(attrs[b2 * 15] * q_scale)
                                scanline.append(attrs[b2 * 12 + b1 * 3])
                                scanline.append(attrs[b1 * 15] * q_scale)
                                scanline.append(attrs[b1 * 8 + b0 * 7])
                                scanline.append(attrs[b0 * 15] * q_scale)
                img_data_blocks.append(compressor.compress(''.join(scanline) * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd1_nt_nf_1udg(self, udg_array, scale, attr_map, trans, flash, x=0, y=0, width=0, height=0, bit_depth=0):
        # Build image data with bit depth 1 for a single UDG
        udg = udg_array[0][0]
        chars = {}
        xor = 255 * attr_map.values()[0][0]
        if scale == 4:
            for b in set(udg.data):
                b7, b6, b5, b4, b3, b2, b1, b0 = BITS8[b ^ xor]
                chars[b] = chr(b7 * 240 + b6 * 15) + chr(b5 * 240 + b4 * 15) + chr(b3 * 240 + b2 * 15) + chr(b1 * 240 + b0 * 15)
        elif scale == 2:
            for b in set(udg.data):
                b7, b6, b5, b4, b3, b2, b1, b0 = BITS8[b ^ xor]
                chars[b] = chr(b7 * 192 + b6 * 48 + b5 * 12 + b4 * 3) + chr(b3 * 192 + b2 * 48 + b1 * 12 + b0 * 3)
        elif scale == 1:
            for b in set(udg.data):
                chars[b] = chr(b ^ xor)
        elif scale == 8:
            for b in set(udg.data):
                b7, b6, b5, b4, b3, b2, b1, b0 = BITS8[b ^ xor]
                chars[b] = chr(b7 * 255) + chr(b6 * 255) + chr(b5 * 255) + chr(b4 * 255) + chr(b3 * 255) + chr(b2 * 255) + chr(b1 * 255) + chr(b0 * 255)

        img_data = ''
        for b in udg.data:
            img_data += (chr(0) + chars[b]) * scale
        return zlib.compress(img_data, self.options[PNG_COMPRESSION_LEVEL])

    def _build_image_data_bd1(self, udg_array, scale, attr_map, trans, flash, x=0, y=0, width=0, height=0, bit_depth=0):
        # Build image data with bit depth 1 and a 2-colour palette
        attr_dict = {}
        for attr, q in attr_map.items():
            attr_dict[attr] = [chr(q[b7] * 128 + q[b6] * 64 + q[b5] * 32 + q[b4] * 16 + q[b3] * 8 + q[b2] * 4 + q[b1] * 2 + q[b0]) for b7, b6, b5, b4, b3, b2, b1, b0 in BITS8]

        compressor = zlib.compressobj(self.options[PNG_COMPRESSION_LEVEL])
        e_scale = scale // 8
        scale_m = scale & 7
        if trans:
            t = (0,) * scale
            ink = paper = (1,) * scale
        img_data_blocks = []
        for row in udg_array:
            for i in range(8):
                scanline = [chr(0)]
                for udg in row:
                    byte = udg.data[i]
                    if trans and udg.mask:
                        # Apply mask
                        p = []
                        mask_byte = udg.mask[i]
                        byte &= mask_byte
                        for b in range(8):
                            if byte & 128:
                                p.extend(ink)
                            elif mask_byte & 128:
                                p.extend(t)
                            else:
                                p.extend(paper)
                            byte *= 2
                            mask_byte *= 2
                        scanline.extend([chr(p[j] * 128 + p[j + 1] * 64 + p[j + 2] * 32 + p[j + 3] * 16 + p[j + 4] * 8 + p[j + 5] * 4 + p[j + 6] * 2 + p[j + 7]) for j in range(0, len(p), 8)])
                    else:
                        # No mask
                        attr = udg.attr
                        if flash and attr & 128:
                            byte ^= 255
                        b7, b6, b5, b4, b3, b2, b1, b0 = BITS8[byte]
                        attrs = attr_dict[attr & 127]
                        if scale_m == 0:
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 1:
                            if scale == 1:
                                scanline.append(attrs[byte])
                                continue
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 128 + b6 * 127])
                            scanline.append(attrs[b6 * 255] * (e_scale - 1))
                            scanline.append(attrs[b6 * 192 + b5 * 63])
                            scanline.append(attrs[b5 * 255] * (e_scale - 1))
                            scanline.append(attrs[b5 * 224 + b4 * 31])
                            scanline.append(attrs[b4 * 255] * (e_scale - 1))
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.append(attrs[b3 * 255] * (e_scale - 1))
                            scanline.append(attrs[b3 * 248 + b2 * 7])
                            scanline.append(attrs[b2 * 255] * (e_scale - 1))
                            scanline.append(attrs[b2 * 252 + b1 * 3])
                            scanline.append(attrs[b1 * 255] * (e_scale - 1))
                            scanline.append(attrs[b1 * 254 + b0])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 2:
                            if scale == 2:
                                scanline.append(attrs[b7 * 192 + b6 * 48 + b5 * 12 + b4 * 3])
                                scanline.append(attrs[b3 * 192 + b2 * 48 + b1 * 12 + b0 * 3])
                                continue
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 192 + b6 * 63])
                            scanline.append(attrs[b6 * 255] * (e_scale - 1))
                            scanline.append(attrs[b6 * 240 + b5 * 15])
                            scanline.append(attrs[b5 * 255] * (e_scale - 1))
                            scanline.append(attrs[b5 * 252 + b4 * 3])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 192 + b2 * 63])
                            scanline.append(attrs[b2 * 255] * (e_scale - 1))
                            scanline.append(attrs[b2 * 240 + b1 * 15])
                            scanline.append(attrs[b1 * 255] * (e_scale - 1))
                            scanline.append(attrs[b1 * 252 + b0 * 3])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 3:
                            if scale == 3:
                                scanline.append(attrs[b7 * 224 + b6 * 28 + b5 * 3])
                                scanline.append(attrs[b5 * 128 + b4 * 112 + b3 * 14 + b2])
                                scanline.append(attrs[b2 * 192 + b1 * 56 + b0 * 7])
                                continue
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 224 + b6 * 31])
                            scanline.append(attrs[b6 * 255] * (e_scale - 1))
                            scanline.append(attrs[b6 * 252 + b5 * 3])
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b5 * 128 + b4 * 127])
                            scanline.append(attrs[b4 * 255] * (e_scale - 1))
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.append(attrs[b3 * 255] * (e_scale - 1))
                            scanline.append(attrs[b3 * 254 + b2])
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b2 * 192 + b1 * 63])
                            scanline.append(attrs[b1 * 255] * (e_scale - 1))
                            scanline.append(attrs[b1 * 248 + b0 * 7])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 4:
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 240 + b6 * 15])
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b5 * 240 + b4 * 15])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 240 + b2 * 15])
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b1 * 240 + b0 * 15])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 5:
                            if scale == 5:
                                scanline.append(attrs[b7 * 248 + b6 * 7])
                                scanline.append(attrs[b6 * 192 + b5 * 62 + b4])
                                scanline.append(attrs[b4 * 240 + b3 * 15])
                                scanline.append(attrs[b3 * 128 + b2 * 124 + b1 * 3])
                                scanline.append(attrs[b1 * 224 + b0 * 31])
                                continue
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 248 + b6 * 7])
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b6 * 192 + b5 * 63])
                            scanline.append(attrs[b5 * 255] * (e_scale - 1))
                            scanline.append(attrs[b5 * 254 + b4])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 128 + b2 * 127])
                            scanline.append(attrs[b2 * 255] * (e_scale - 1))
                            scanline.append(attrs[b2 * 252 + b1 * 3])
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b1 * 224 + b0 * 31])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 6:
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 252 + b6 * 3])
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b6 * 240 + b5 * 15])
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b5 * 192 + b4 * 63])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 252 + b2 * 3])
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b2 * 240 + b1 * 15])
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b1 * 192 + b0 * 63])
                            scanline.append(attrs[b0 * 255] * e_scale)
                        elif scale_m == 7:
                            scanline.append(attrs[b7 * 255] * e_scale)
                            scanline.append(attrs[b7 * 254 + b6])
                            scanline.append(attrs[b6 * 255] * e_scale)
                            scanline.append(attrs[b6 * 252 + b5 * 3])
                            scanline.append(attrs[b5 * 255] * e_scale)
                            scanline.append(attrs[b5 * 248 + b4 * 7])
                            scanline.append(attrs[b4 * 255] * e_scale)
                            scanline.append(attrs[b4 * 240 + b3 * 15])
                            scanline.append(attrs[b3 * 255] * e_scale)
                            scanline.append(attrs[b3 * 224 + b2 * 31])
                            scanline.append(attrs[b2 * 255] * e_scale)
                            scanline.append(attrs[b2 * 192 + b1 * 63])
                            scanline.append(attrs[b1 * 255] * e_scale)
                            scanline.append(attrs[b1 * 128 + b0 * 127])
                            scanline.append(attrs[b0 * 255] * e_scale)
                img_data_blocks.append(compressor.compress(''.join(scanline) * scale))
        img_data_blocks.append(compressor.flush())
        return ''.join(img_data_blocks)

    def _build_image_data_bd1_blank(self, udg_array, scale, attr_map=None, trans=0, flash=0, x=0, y=0, width=0, height=0, bit_depth=0):
        # Build image data with bit depth 1 and a single-colour palette;
        # placing the integer value in brackets means it is evaluated before
        # 'multiplying' the string (and is therefore quicker)
        scanlines = chr(0) * ((1 + len(udg_array[0]) * scale) * len(udg_array) * scale * 8)
        return zlib.compress(scanlines, self.options[PNG_COMPRESSION_LEVEL])
